@@ -3,19 +3,19 @@
 %%
 %%     Segmentation UI for AutoUV.
 %%
-%%  Copyright (c) 2002-2004 Dan Gudmundsson, Bjorn Gustavsson
+%%  Copyright (c) 2002-2011 Dan Gudmundsson, Bjorn Gustavsson
 %%
 %%  See the file "license.terms" for information on usage and redistribution
 %%  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 %%
-%%     $Id: auv_seg_ui.erl 391 2008-10-03 00:23:36Z optigon $
+%%     $Id$
 %%
 
 -module(auv_seg_ui).
 -export([start/4]).
 
 -define(NEED_ESDL, 1).
--include("wings.hrl").
+-include_lib("wings/src/wings.hrl").
 -include("auv.hrl").
 
 -import(lists, [sort/1,map/2,member/2,foldl/3,reverse/1]).
@@ -36,11 +36,6 @@
 start(Op, #we{id=Id}=We0, OrigWe, St0) ->
     Modes = [vertex,edge,face],
     wings:mode_restriction(Modes),
-    This = wings_wm:this(),
-    Allowed = [view,select],
-    Menu  = [Item || {_,Name,_}=Item <- get(wings_menu_template),
-		     member(Name, Allowed)],
-    wings_wm:menubar(This, Menu),
     wings_pref:set_value(scene_lights, false), %% ugly hack 
     
     St1 = seg_create_materials(St0),
@@ -54,9 +49,9 @@ start(Op, #we{id=Id}=We0, OrigWe, St0) ->
 				   Es = wings_face:outer_edges(Chart,We1),
 				   gb_sets:union(gb_sets:from_list(Es),Acc)
 			   end, Cuts0, Charts),
-		 We1#we{he=Cuts, mode=material};
+		 We1#we{he=Cuts};
 	     delete_old ->
-		 We1#we{mode=material}
+		 We1
 	 end,
 
     St = St1#st{sel=[],selmode=face,shapes=gb_trees:from_orddict([{Id,We}])},
@@ -209,17 +204,7 @@ seg_event_6({action,{view,Cmd}}, #seg{st=St0}=Ss) ->
             wings_view:command(aim, St1),
             get_seg_event(Ss);
           false ->
-            HL0 = wings_pref:get_value(highlight_aim_at_unselected),
-            HL1 = wings_pref:get_value(highlight_aim_at_selected),
-            {_,X,Y} = wings_wm:local_mouse_state(),
-            {{_,Cmd0},St1} = case wings_pick:do_pick(X, Y, St0) of
-                  {add,_,St2} when HL0 =:= true ->
-                      {{view,highlight_aim},{add,St0,St2}};
-                  {delete,_,St2} when HL1 =:= true ->
-                      {{view,highlight_aim},{delete,St0,St2}};
-                  _Other -> 
-                      {{view,aim}, St0}
-            end,
+            {{_,Cmd0},St1} = wings:highlight_aim_setup(St0),
             wings_view:command(Cmd0,St1),
             get_seg_event(Ss)
         end;
@@ -288,13 +273,13 @@ seg_command(select_hard_edges, #seg{st=St0}=Ss) ->
 	#st{}=St -> get_seg_event(Ss#seg{st=St})
     end;
 seg_command({select,Mat}, #seg{st=St0}=Ss) ->
-    St = wings_material:command({select,[atom_to_list(Mat)]}, St0),
+    {save_state,St} = wings_material:command({select,[atom_to_list(Mat),select]}, St0),
     get_seg_event(Ss#seg{st=St});
 seg_command({segment,Type}, #seg{st=St0}=Ss) ->
     St = segment(Type, St0),
     get_seg_event(Ss#seg{st=St});
 seg_command({debug,Cmd}, Ss) ->
-    seg_command_debug(Cmd, Ss);
+    seg_command_debug({debug,Cmd}, Ss);
 seg_command(ignore_faces,#seg{st=St0,fs=Mode0}=Ss) ->    
     HiddenFs = wpa:sel_fold(fun(Fs,_,Acc) -> gb_sets:union(Fs,Acc) end,
 			    gb_sets:empty(), St0),
@@ -323,12 +308,12 @@ seg_command(Cmd, #seg{st=#st{mat=Mat}=St0}=Ss) ->
 seg_command_debug(_, _) -> keep.
 -else.
 seg_command_debug({debug,select_features}, #seg{we=#we{id=Id}=We,st=St}=Ss) ->
-    Tot = gb_trees:size(We#we.es),
+    Tot = array:sparse_size(We#we.es),
     {Es,_,_} = auv_segment:find_features(We, 60, Tot div 50),
     Sel = [{Id,gb_sets:from_list(Es)}],
     get_seg_event(Ss#seg{st=St#st{selmode=edge,sel=Sel}});
 seg_command_debug({debug,select_seeds}, #seg{we=#we{id=Id}=We,st=St}=Ss) ->
-    Tot = gb_trees:size(We#we.es),
+    Tot = array:sparse_size(We#we.es),
     {Features,_,_} = auv_segment:find_features(We, 60, Tot div 50),
     {Seeds0,_} = auv_segment:build_seeds(Features, We),
     Seeds = [S || {_,S} <- Seeds0],
@@ -376,7 +361,7 @@ seg_map_charts(Method, #seg{st=#st{shapes=Shs},we=OrigWe}=Ss) ->
     wings_pb:done(),
     case length(Charts) of
 	0 ->
-	    wings_u:error(?__(5,"No mappable faces."));
+	    wings_u:error_msg(?__(5,"No mappable faces."));
 	N ->
 	    wings_pb:start(?__(6,"mapping")),
 	    seg_map_charts_1(Charts, Method, 1, N, [], [], Ss)
@@ -390,7 +375,7 @@ seg_map_charts_1([We0|Cs], Type, Id, N, Acc,Failed,Ss) ->
 	    seg_map_charts_1(Cs,Type,Id+1,N,Acc,[We1|Failed],Ss#seg{err=Message});
 	Vs0 ->
 	    Vs = auv_placement:rotate_area(Vs0,We1),
-	    We = We1#we{vp=gb_trees:from_orddict(sort(Vs))},
+	    We = We1#we{vp=array:from_orddict(sort(Vs))},
 	    seg_map_charts_1(Cs, Type, Id+1, N, [We|Acc], Failed, Ss)
     end;
 seg_map_charts_1([],_, _, _,Charts0,Failed,
@@ -448,12 +433,8 @@ fake_selection(St) ->
 		  end, St).
 
 fake_sel_1(St0) ->
-    case wings_pref:get_value(use_temp_sel) of
-	false -> St0;
-	true ->
-	    {_,X,Y} = wings_wm:local_mouse_state(),
-	    case wings_pick:do_pick(X, Y, St0) of
-		{add,_,St} -> St;
-		_ -> St0
-	    end
+    {_,X,Y} = wings_wm:local_mouse_state(),
+    case wings_pick:do_pick(X, Y, St0) of
+	{add,_,St} -> St;
+	_ -> St0
     end.
