@@ -3,34 +3,34 @@
 %%
 %%     This module handles camera moves (rotation, zooming, and panning).
 %%
-%%  Copyright (c) 2001-2004 Bjorn Gustavsson
+%%  Copyright (c) 2001-2011 Bjorn Gustavsson
 %%
 %%  See the file "license.terms" for information on usage and redistribution
 %%  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 %%
-%%     $Id: wings_camera.erl 376 2008-09-15 22:20:34Z optigon $
+%%     $Id$
 %%
 
 -module(wings_camera).
 -export([init/0,prefs/0,help/0,event/2,event/3]).
 
+-export([tweak_camera_event/4]).
+
 -define(NEED_ESDL, 1).
 -include("wings.hrl").
 
--import(lists, [foreach/2,map/2,foldl/3,sort/1,reverse/1,append/1]).
-
 -define(ZOOM_FACTOR, 20).
 -define(ZOOM_FACTOR_ALT, 1).
 
 -define(CAMDIV, 4).
--define(CAMMAX, 150).  %% Always larger than 300 on my pc
 
 -define(CSEP, 160).				%Short space.
 -define(NEARZERO, 1.0e-6).
 -record(camera,
 	{x,y,					%Current mouse position.
 	 ox,oy,					%Original mouse position.
-	 xt=0,yt=0				%Last warp length.
+	 xt=0,yt=0,				%Last warp length.
+         w,h                                    %Win size
 	}).
 
 -record(state, {st, func}).
@@ -40,13 +40,12 @@ init() ->
     wings_pref:set_default(num_buttons, 3),
     wings_pref:set_default(cam_rotation_speed, 25),
     wings_pref:set_default(pan_speed, 25),
-    wings_pref:set_default(pan_using_arrow_keys, true),
     wings_pref:set_default(pan_speed_arrow_keys, 50),
-    wings_pref:set_default(inverted_wheel_zoom, false),
     wings_pref:set_default(wheel_adds, false),
     wings_pref:set_default(wh_scroll_info,true),
     wings_pref:set_default(wh_pan_spd, 50),
     wings_pref:set_default(wh_rot_spd, 7.50),
+    wings_pref:set_default(highlight_zoom_aim, false),
     case {wings_pref:get_value(num_buttons),wings_pref:get_value(camera_mode)} of
 	{3,_} -> ok;
 	{_,nendo} -> ok;
@@ -61,70 +60,62 @@ prefs() ->
     CamRotSpeed = wings_pref:get_value(cam_rotation_speed, 25),
     PanSpeed0 = wings_pref:get_value(pan_speed, 25),
     ArrowPanSpeed = wings_pref:get_value(pan_speed_arrow_keys, 50),
-    InvertZW = wings_pref:get_value(inverted_wheel_zoom, false),
-    WheelAdds = wings_pref:get_value(wheel_adds,false),
+    WheelAdds = wings_pref:get_value(wheel_adds,true),
     WhScrollInfo = wings_pref:get_value(wh_scroll_info,true),
-    ArrowKeysPan = wings_pref:get_value(pan_using_arrow_keys,true),
     WhPanSpd = wings_pref:get_value(wh_pan_spd, 50),
     WhRotate = wings_pref:get_value(wh_rot_spd, 7.5),
+    ZoomAim = wings_pref:get_value(highlight_zoom_aim),
+
+    AddHook = fun(wheel_adds, Enable0, Sto) ->
+		      Keys = [wh_scroll_info, wh_pan_spd, wh_rot_spd],
+		      Enable = Enable0 andalso wings_dialog:get_value(wheel_zooms, Sto),
+		      wings_dialog:enable(Keys, Enable, Sto)
+	      end,
+    WheelHook = fun(wheel_zooms, Enable, Sto) ->
+			Keys = [wheel_zoom_factor,wheel_zoom_factor_alt,wheel_adds],
+			wings_dialog:enable(Keys, Enable, Sto),
+			Keys2 = [wh_scroll_info, wh_pan_spd, wh_rot_spd],
+			Enable2 = Enable andalso wings_dialog:get_value(wheel_adds, Sto),
+			wings_dialog:enable(Keys2, Enable2, Sto)
+		end,
 
-    Hook = fun (is_disabled, {_Var,_I,Sto}) ->
-            not gb_trees:get(wheel_zooms, Sto);
-             (_, _) -> void
-           end,
-    WHook = fun (is_disabled, {_Var,_I,Sto}) ->
-             not ((gb_trees:get(wheel_adds, Sto)) andalso (gb_trees:get(wheel_zooms, Sto)));
-             (_, _) -> void
-            end,
-    ArrowKeyHook = fun (is_disabled, {_Var,_I,Sto}) ->
-            not gb_trees:get(pan_using_arrow_keys, Sto);
-             (_, _) -> void
-           end,
     {hframe,
-    [{vframe,
-     [{vframe,[mouse_buttons()],[{title,?__(1,"Mouse Buttons")}]},
-      {vframe,[camera_modes()],[{title,?__(2,"Camera Mode")}]},
+     [{vframe,
+       [{vframe,[mouse_buttons()],[{title,?__(1,"Mouse Buttons")}]},
+	{vframe,[camera_modes()],[{title,?__(2,"Camera Mode")}]},
+	{vframe,
+	 [{hframe,[{slider,{text,CamRotSpeed,[{key,cam_rotation_speed},{range,{1,100}}]}}]}],
+	 [{title,?__(19,"Rotation Speed")}]},
+	{vframe,
+	 [{hframe,[{slider,{text,PanSpeed0,[{key,pan_speed},{range,{1,100}}]}}]}],
+	 [{title,?__(3,"Pan Speed")}]},
+	{vframe,
+	 [{hframe,[{slider,{text,ArrowPanSpeed,[{key,pan_speed_arrow_keys},{range,{1,100}}]}}]}],
+	 [{title,?__(16,"Arrow Key Pan Speed")}]}]},
       {vframe,
-       [{hframe,[{slider,{text,CamRotSpeed,[{key,cam_rotation_speed},{range,{1,100}}]}}]}],
-       [{title,?__(19,"Rotation Speed")}]},
-      {vframe,
-       [{hframe,[{slider,{text,PanSpeed0,[{key,pan_speed},{range,{1,100}}]}}]}],
-       [{title,?__(3,"Pan Speed")}]},
-      {vframe,
-       [{?__(18,"Arrow Keys Pan"),ArrowKeysPan,[{key,pan_using_arrow_keys}]},
-       {hframe,[{slider,{text,ArrowPanSpeed,[{key,pan_speed_arrow_keys},{range,{1,100}},
-        {hook,ArrowKeyHook}]}}]}],
-       [{title,?__(16,"Arrow Key Pan Speed")}]}]},
-    {vframe,
-      [{vframe,
-       [{?__(4,"Wheel Zooms"),ZoomFlag0,[{key,wheel_zooms}]},
-    {vradio,[{?__(5,"Forwards Zooms In"),false},
-             {?__(6,"Forwards Zooms Out"),true}],
-     InvertZW,
-     [{key,inverted_wheel_zoom},{hook,Hook}]},
-    {hframe,
-     [{label,?__(7,"Zoom Factor"),[{hook,Hook}]},
-      {text,ZoomFactor0,
-       [{key,wheel_zoom_factor},
-     {range,{1,50}},{hook,Hook}]},
-      {label,"%",[{hook,Hook}]}]},
-     {hframe,
-      [{label,?__(10,"Alternate Zoom Factor(Alt+Scroll)"),[{hook,Hook}]},
-       {text,ZoomFactorAlt,
-        [{key,wheel_zoom_factor_alt},
-      {range,{1,50}},{hook,Hook}]},
-     {label,"%",[{hook,Hook}]}]} ],[{title,?__(9,"Scroll Wheel")}]},
-    {vframe,
-      [{?__(11,"Wheel Pans & Rotates"),WheelAdds,[{key,wheel_adds},{hook,Hook}]},
-       {?__(17,"Show Info Line Help String"),WhScrollInfo,[{key,wh_scroll_info},{hook,WHook}]}, 
-      {vframe,
-       [{hframe,
-        [{slider,{text,WhPanSpd,[{key, wh_pan_spd},{range,{1,100}}]}}],
-      [{title,?__(12,"Pan Speed")},{hook,WHook}]},
-      {hframe,
-       [{slider,{text,WhRotate,[{key, wh_rot_spd},{range,{?NEARZERO,180.0}}]}}],
-       [{title,?__(13,"Rotation Step in Degrees")},{hook,WHook}]}]}],
-    [{title,?__(15,"Unidirectional Camera")}]}]}]}.
+       [{vframe,
+	 [{hframe,[{?__(4,"Wheel Zooms"),ZoomFlag0,[{key,wheel_zooms}, {hook, WheelHook}]},
+		   {?__(20,"Zooming in aims Camera"),ZoomAim,[{key,highlight_zoom_aim}]}]},
+	  {hframe,
+	   [{label,?__(7,"Zoom Factor")},
+	    {text,ZoomFactor0, [{key,wheel_zoom_factor},{range,{1,50}}]},
+	    {label,"%"}]},
+	  {hframe,
+	   [{label,?__(10,"Alternate Zoom Factor(Alt+Scroll)")},
+	    {text,ZoomFactorAlt, [{key,wheel_zoom_factor_alt},{range,{1,50}}]},
+	    {label,"%"}]}
+	 ],[{title,?__(9,"Scroll Wheel")}]},
+	{vframe,
+	 [{?__(11,"Wheel Pans and Rotates"),WheelAdds,[{key,wheel_adds}, {hook, AddHook}]},
+	  {?__(17,"Show Info Line Help String"),WhScrollInfo,[{key,wh_scroll_info}]},
+	  {vframe,
+	   [{hframe,
+	     [{slider,{text,WhPanSpd,[{key, wh_pan_spd},{range,{1,100}}]}}],
+	     [{title,?__(12,"Pan Speed")}]},
+	    {hframe,
+	     [{slider,{text,WhRotate,[{key, wh_rot_spd},{range,{?NEARZERO,180.0}}]}}],
+	     [{title,?__(13,"Rotation Step in Degrees")}]}]}],
+	 [{title,?__(15,"Unidirectional Camera")}]}]}]}.
 
 mouse_buttons() ->
     {menu,[{desc(1),1,[{info,info(1)}]},
@@ -132,32 +123,29 @@ mouse_buttons() ->
            {desc(3),3,[{info,info(3)}]}],
      wings_pref:get_value(num_buttons),
      [{key,num_buttons},
-      {hook,fun (update, {Var,_I,Val,Sto0}) ->
-		    Sto = gb_trees:update(Var, Val, Sto0),
-		    Mode0 = gb_trees:get(camera_mode, Sto),
-		    Mode = case {Val,Mode0} of
-			       {1,_} -> nendo;
-			       {2,blender} -> blender;
-			       {2,_} -> nendo;
-			       {3,_} -> Mode0
-			   end,
-		    {store,gb_trees:update(camera_mode, Mode, Sto)};
-		(_, _) -> void
+      {hook,fun (_,Val,Sto) ->
+		    Mode0 = wings_dialog:get_value(camera_mode, Sto),
+		    Mode  = case {Val,Mode0} of
+				{1,_} -> nendo;
+				{2,blender} -> blender;
+				{2,_} -> nendo;
+				{3,_} -> Mode0
+			    end,
+		    wings_dialog:set_value(camera_mode, {Mode, camera_modes(Val)}, Sto)
 	    end}]}.
 
 camera_modes() ->
-    Modes = [wings_cam,mirai,nendo,maya,tds,blender,mb,sketchup],
-    {menu,[{desc(Mode),Mode,[{info,info(Mode)}]} || Mode <- Modes],
+    {menu,camera_modes(wings_pref:get_value(num_buttons)),
      wings_pref:get_value(camera_mode),
-     [{key,camera_mode},
-      {hook,fun (menu_disabled, {_Var,_I,Sto}) ->
-		    case gb_trees:get(num_buttons, Sto) of
-			1 -> [mirai,maya,tds,blender,mb,sketchup,wings_cam];
-			2 -> [mirai,maya,tds,mb,sketchup,wings_cam];
-			3 -> []
-		    end;
-		(_, _) -> void
-	    end}]}.
+     [{key,camera_mode}]}.
+
+camera_modes(NumButtons) ->
+    All = [wings_cam,mirai,nendo,maya,tds,blender,mb,sketchup],
+    [{desc(Mode),Mode,[{info,info(Mode)}]} || Mode <- All -- disabled(NumButtons)].
+
+disabled(1) -> [mirai,maya,tds,blender,mb,sketchup,wings_cam];
+disabled(2) -> [mirai,maya,tds,mb,sketchup,wings_cam];
+disabled(3) -> [].
 
 desc(1) -> ?__(1,"One");
 desc(2) -> ?__(2,"Two");
@@ -200,65 +188,83 @@ scroll_help() ->
 %% Event handler.
 event(Ev, St=#st{}) -> 
     event(Ev,St,none).
-%% Rotation event
-event(#mousebutton{button=4,mod=Mod,state=?SDL_RELEASED}, _, _Redraw)
-  when Mod band ?SHIFT_BITS =/= 0 andalso Mod band ?ALT_BITS =/= 0 ->
-    whrotate(0.5,0.0);
-event(#mousebutton{button=5,mod=Mod,state=?SDL_RELEASED}, _, _Redraw)
-  when Mod band ?SHIFT_BITS =/= 0 andalso Mod band ?ALT_BITS =/= 0 ->
-    whrotate(-0.5,0.0);
-event(#mousebutton{button=4,mod=Mod,state=?SDL_RELEASED}, _, _Redraw)
-  when Mod band ?SHIFT_BITS =/= 0 ->
-    whrotate(0.0,0.5);
-event(#mousebutton{button=5,mod=Mod,state=?SDL_RELEASED}, _, _Redraw)
-  when Mod band ?SHIFT_BITS =/= 0 ->
-    whrotate(0.0,-0.5);
-%% Pan event
-event(#mousebutton{button=4,mod=Mod,state=?SDL_RELEASED}, _, _Redraw)
-  when Mod band ?CTRL_BITS =/= 0 andalso Mod band ?ALT_BITS =/= 0 ->
-    whpan(0.5,0.0);
-event(#mousebutton{button=5,mod=Mod,state=?SDL_RELEASED}, _, _Redraw)
-  when Mod band ?CTRL_BITS =/= 0 andalso Mod band ?ALT_BITS =/= 0 ->
-    whpan(-0.5,0.0);
-event(#mousebutton{button=4,mod=Mod,state=?SDL_RELEASED}, _, _Redraw)
-  when Mod band ?CTRL_BITS =/= 0 ->
-    whpan(0.0,0.5);
-event(#mousebutton{button=5,mod=Mod,state=?SDL_RELEASED}, _, _Redraw)
-  when Mod band ?CTRL_BITS =/= 0 ->
-    whpan(0.0,-0.5);
-%% Zoom event
-event(#mousebutton{button=4,mod=Mod,state=?SDL_RELEASED}, _, _Redraw)
-  when Mod band ?ALT_BITS =/= 0 ->
-    zoom_step_alt(-1);
-event(#mousebutton{button=4,state=?SDL_RELEASED}, _, _Redraw) ->
-    zoom_step(-1);
-event(#mousebutton{button=5,mod=Mod,state=?SDL_RELEASED}, _, _Redraw)
-  when Mod band ?ALT_BITS =/= 0 ->
-    zoom_step_alt(1);
-event(#mousebutton{button=5,state=?SDL_RELEASED}, _, _Redraw) ->
-    zoom_step(1);
-event(#mousebutton{button=B}, _, _Redraw) when B==4; B==5 ->
-    keep;
+%% Scroll wheel camera events
+event(#mousewheel{}=Ev, _St, _Redraw) ->
+    generic_event(Ev,_St,_Redraw);
+% Camera mode specific events
 event(Ev, St, Redraw) ->
     case wings_pref:get_value(camera_mode) of
+	wings_cam -> wings_cam(Ev, #state{st=St, func=Redraw});
 	blender -> blender(Ev, #state{st=St, func=Redraw});
 	nendo -> nendo(Ev, #state{st=St, func=Redraw});
 	mirai -> mirai(Ev, #state{st=St, func=Redraw});
 	tds -> tds(Ev, #state{st=St, func=Redraw});
 	maya -> maya(Ev, #state{st=St, func=Redraw});
 	mb -> mb(Ev, #state{st=St, func=Redraw});
-	sketchup -> sketchup(Ev, #state{st=St, func=Redraw});
-	wings_cam -> wings_cam(Ev, #state{st=St, func=Redraw})
+	sketchup -> sketchup(Ev, #state{st=St, func=Redraw})
+    end.
+
+%%%
+%%% Tweak Camera
+%%%
+
+tweak_camera_event(Sym, X, Y, St) when Sym =:= $c; Sym =:= $s; Sym =:= $d ->
+    Camera = start_camera(X,Y),
+    {seq,push,get_tweak_cam_event(Sym, Camera, St)};
+tweak_camera_event(Sym, _, _, _) ->
+    arrow_key_pan(Sym).
+
+get_tweak_cam_event(Sym, Camera, St) ->
+    {replace,fun(Ev) -> tweak_cam_event(Ev, Sym, Camera, St) end}.
+
+tweak_cam_event(#mousemotion{x=X0,y=Y0}, Sym, Camera0, St) ->
+    case wings_io:is_key_pressed(Sym) of
+      true  ->
+          {Dx,Dy,Camera} = camera_mouse_range(X0, Y0, Camera0),
+          case Sym of
+              $c -> rotate(Dx,Dy);
+              $s -> pan(Dx,Dy);
+              $d -> zoom(Dy)
+          end,
+          wings_wm:dirty(),
+          get_tweak_cam_event(Sym, Camera, St);
+      false ->
+        quit_tweak_cam()
+     end;
+tweak_cam_event(#mousebutton{button=B,state=?SDL_RELEASED}=Ev, Sym, Camera, St)
+  when B =< 3->
+    case wings_io:is_key_pressed(Sym) of
+      true ->
+        generic_event(Ev, Camera, #state{st=St, func=none});
+      false ->
+        quit_tweak_cam()
+     end;
+tweak_cam_event(Ev, Sym, Camera, St) ->
+    case wings_io:is_key_pressed(Sym) of
+      true ->
+        generic_event(Ev, Camera, #state{st=St, func=none});
+      false ->
+        quit_tweak_cam()
+     end.
+
+quit_tweak_cam() ->
+    case wings_io:get_mouse_state() of
+	{0,X0,Y0} ->
+	    {X,Y} = wings_wm:screen2local({X0,Y0}),
+	    wings_wm:later(#mousebutton{which=wings_wm:this_win(),
+					button=1,x=X,y=Y,mod=0,state=?SDL_RELEASED}),
+	    pop;
+	_ ->
+	    pop
     end.
 
 %%%
 %%% Blender style camera.
 %%%
 
-blender(#mousebutton{button=2,state=?SDL_PRESSED,x=X0,y=Y0,mod=Mod}, Redraw)
+blender(#mousebutton{button=2,state=?SDL_PRESSED,x=X,y=Y,mod=Mod}, Redraw)
   when Mod band ?ALT_BITS =:= 0 ->
-    {X,Y} = wings_wm:local2global(X0, Y0),
-    Camera = #camera{x=X,y=Y,ox=X,oy=Y},
+    Camera = start_camera(X,Y),
     grab(),
     message(blender_help()),
     {seq,push,get_blender_event(Camera, Redraw)};
@@ -298,10 +304,9 @@ blender_help() ->
 %%% Nendo style camera.
 %%%
 
-nendo(#mousebutton{button=2,x=X0,y=Y0,mod=Mod,state=?SDL_RELEASED}, Redraw)
+nendo(#mousebutton{button=2,x=X,y=Y,mod=Mod,state=?SDL_RELEASED}, Redraw)
   when Mod band ?CTRL_BITS =:= 0 ->
-    {X,Y} = wings_wm:local2global(X0, Y0),
-    Camera = #camera{x=X,y=Y,ox=X,oy=Y},
+    Camera = start_camera(X,Y),
     grab(),
     MoveTumbles = allow_rotation(),
     nendo_message(MoveTumbles),
@@ -315,18 +320,18 @@ nendo_event(#mousebutton{button=1,state=?SDL_RELEASED}, Camera, _, _) ->
 nendo_event(#mousemotion{x=X,y=Y,state=Buttons}, Camera0, Redraw, true) ->
     {Dx,Dy,Camera} = camera_mouse_range(X, Y, Camera0),
     case Buttons band 6 of
-	0 ->					%None of MMB/RMB pressed.
+	0 ->
 	    rotate(-Dx, -Dy);
-	_Other ->				%MMB and/or RMB pressed.
+	_Other ->
 	    zoom(Dy)
     end,
     get_nendo_event(Camera, Redraw, true);
 nendo_event(#mousemotion{x=X,y=Y,state=Buttons}, Camera0, Redraw, false) ->
     {Dx,Dy,Camera} = camera_mouse_range(X, Y, Camera0),
     case Buttons band 6 of
-	0 ->					%None of MMB/RMB pressed.
+	0 ->
 	    pan(-Dx, -Dy);
-	_Other ->				%MMB and/or RMB pressed.
+	_Other ->
 	    zoom(Dy)
     end,
     get_nendo_event(Camera, Redraw, false);
@@ -368,10 +373,9 @@ nendo_message(false) ->
 %%% Mirai style camera.
 %%%
 
-mirai(#mousebutton{button=2,x=X0,y=Y0,mod=Mod,state=?SDL_RELEASED}, Redraw)
+mirai(#mousebutton{button=2,x=X,y=Y,mod=Mod,state=?SDL_RELEASED}, Redraw)
   when Mod band ?CTRL_BITS =:= 0 ->
-    {X,Y} = wings_wm:local2global(X0, Y0),
-    Camera = #camera{x=X,y=Y,ox=X,oy=Y},
+    Camera = start_camera(X,Y),
     grab(),
     MoveTumbles = allow_rotation(),
     mirai_message(MoveTumbles),
@@ -443,8 +447,7 @@ mirai_message(false) ->
 %%% 3ds max style camera.
 %%%
 
-tds(#mousebutton{button=2,x=X0,y=Y0,state=?SDL_PRESSED}, Redraw) ->
-    {X,Y} = wings_wm:local2global(X0, Y0),
+tds(#mousebutton{button=2,x=X,y=Y,state=?SDL_PRESSED}, Redraw) ->
     Camera = #camera{x=X,y=Y,ox=X,oy=Y},
     grab(),
     message(wings_msg:join(tds_help(), 
@@ -493,10 +496,9 @@ tds_help() ->
 %%% Maya style camera.
 %%%
 
-maya(#mousebutton{x=X0,y=Y0,mod=Mod,state=?SDL_PRESSED}, Redraw)
-  when Mod band ?ALT_BITS =/= 0 ->
-    {X,Y} = wings_wm:local2global(X0, Y0),
-    sdl_events:eventState(?SDL_KEYUP, ?SDL_ENABLE),
+maya(#mousebutton{x=X,y=Y,mod=Mod,state=?SDL_PRESSED}, Redraw)
+  when Mod band ?ALT_BITS =/= 0, Mod band ?CTRL_BITS =:= 0, Mod band ?SHIFT_BITS =:= 0 ->
+    wings_io:change_event_handler(?SDL_KEYUP, true),
     Camera = #camera{x=X,y=Y,ox=X,oy=Y},
     grab(),
     message(maya_help()),
@@ -506,11 +508,11 @@ maya(#keyboard{sym=Sym}, _Redraw) ->
 maya(_, _) -> next.
 
 maya_event(#keyboard{sym=Alt,state=?SDL_RELEASED},
-	   Camera, _Redraw) when Alt == ?SDLK_LALT; Alt == ?SDLK_RALT ->
+	   Camera, _Redraw) when Alt =:= ?SDLK_LALT; Alt =:= ?SDLK_RALT ->
     maya_stop_camera(Camera);
 maya_event(#mousebutton{button=B,state=?SDL_RELEASED}, Camera, _)
   when B < 4 ->
-    case sdl_mouse:getMouseState() of
+    case wings_io:get_mouse_state() of
 	{0,_,_} ->
 	    %% Exit camera mode if all mouse buttons released.
 	    maya_stop_camera(Camera);
@@ -521,13 +523,13 @@ maya_event(#mousebutton{button=B,state=?SDL_RELEASED}, Camera, _)
 maya_event(#mousemotion{x=X,y=Y,state=Buttons}, Camera0, Redraw) ->
     {Dx,Dy,Camera} = camera_mouse_range(X, Y, Camera0),
     if
-	Buttons band 4 == 4 ->			%RMB
+	Buttons band 4 =:= 4 ->			%RMB
 	    zoom(-Dx);
-	Buttons band 3 == 3 ->			%LMB+MMB
+	Buttons band 3 =:= 3 ->			%LMB+MMB
 	    zoom(-Dx);
-	Buttons band 1 == 1 ->			%LMB
+	Buttons band 1 =:= 1 ->			%LMB
 	    rotate(Dx, Dy);
-	Buttons band 2 == 2 ->			%MMB
+	Buttons band 2 =:= 2 ->			%MMB
 	    pan(Dx, Dy);
 	true -> ok
     end,
@@ -540,7 +542,7 @@ get_maya_event(Camera, Redraw) ->
     {replace,fun(Ev) -> maya_event(Ev, Camera, Redraw) end}.
 
 maya_stop_camera(Camera) ->
-    sdl_events:eventState(?SDL_KEYUP, ?SDL_IGNORE),
+    wings_io:change_event_handler(?SDL_KEYUP, false),
     stop_camera(Camera).
 
 maya_help() ->
@@ -555,9 +557,8 @@ maya_help() ->
 %%% Motionbuilder style camera.
 %%%
 
-mb(#mousebutton{button=1,mod=Mod,x=X0,y=Y0,state=?SDL_PRESSED}, Redraw)
+mb(#mousebutton{button=1,mod=Mod,x=X,y=Y,state=?SDL_PRESSED}, Redraw)
   when Mod band (?SHIFT_BITS bor ?CTRL_BITS) =/= 0 ->
-    {X,Y} = wings_wm:local2global(X0, Y0),
     Camera = #camera{x=X,y=Y,ox=X,oy=Y},
     grab(),
     message(mb_help()),
@@ -602,9 +603,8 @@ mb_help() ->
 %%% Sketchup style camera.
 %%%
 
-sketchup(#mousebutton{button=2,state=?SDL_PRESSED,x=X0,y=Y0,mod=Mod}, Redraw)
-  when Mod band ?ALT_BITS =:= 0 ->
-    {X,Y} = wings_wm:local2global(X0, Y0),
+sketchup(#mousebutton{button=2,state=?SDL_PRESSED,x=X,y=Y,mod=Mod}, Redraw)
+  when (Mod band (?ALT_BITS  bor ?CTRL_BITS)) =:= 0  ->
     Camera = #camera{x=X,y=Y,ox=X,oy=Y},
     grab(),
     message(sketchup_help()),
@@ -643,10 +643,9 @@ sketchup_help() ->
 %%% Wings 3D camera suggested by oort
 %%%
 
-wings_cam(#mousebutton{button=2,x=X0,y=Y0,mod=Mod,state=?SDL_RELEASED}, Redraw)
-  when Mod band ?CTRL_BITS =:= 0 ->
-    {X,Y} = wings_wm:local2global(X0, Y0),
-    Camera = #camera{x=X,y=Y,ox=X,oy=Y},
+wings_cam(#mousebutton{button=2,x=X,y=Y,mod=Mod,state=?SDL_RELEASED}, Redraw)
+  when Mod band (?CTRL_BITS bor ?SHIFT_BITS bor ?ALT_BITS) =:= 0 ->
+    Camera = start_camera(X,Y),
     grab(),
     wings_cam_message(),
     View = wings_view:current(),
@@ -699,44 +698,68 @@ generic_event(redraw, _Camera, #state{func=Redraw}) when is_function(Redraw) ->
     Redraw(),
     keep;
 
-generic_event(#mousebutton{button=4,mod=Mod,state=?SDL_RELEASED}, _Camera, _Redraw)
-  when Mod band ?SHIFT_BITS =/= 0 andalso Mod band ?ALT_BITS =/= 0 ->
-    whrotate(0.5,0.0);
-generic_event(#mousebutton{button=5,mod=Mod,state=?SDL_RELEASED}, _Camera, _Redraw)
-  when Mod band ?SHIFT_BITS =/= 0 andalso Mod band ?ALT_BITS =/= 0 ->
-    whrotate(-0.5,0.0);
-generic_event(#mousebutton{button=4,mod=Mod,state=?SDL_RELEASED}, _Camera, _Redraw)
-  when Mod band ?SHIFT_BITS =/= 0 ->
-    whrotate(0.0,0.5);
-generic_event(#mousebutton{button=5,mod=Mod,state=?SDL_RELEASED}, _Camera, _Redraw)
-  when Mod band ?SHIFT_BITS =/= 0 ->
-    whrotate(0.0,-0.5);
-
-generic_event(#mousebutton{button=4,mod=Mod,state=?SDL_RELEASED}, _Camera, _Redraw)
-  when Mod band ?CTRL_BITS =/= 0 andalso Mod band ?ALT_BITS =/= 0 ->
-    whpan(0.5,0.0);
-generic_event(#mousebutton{button=5,mod=Mod,state=?SDL_RELEASED}, _Camera, _Redraw)
-  when Mod band ?CTRL_BITS =/= 0 andalso Mod band ?ALT_BITS =/= 0 ->
-    whpan(-0.5,0.0);
-generic_event(#mousebutton{button=4,mod=Mod,state=?SDL_RELEASED}, _Camera, _Redraw)
-  when Mod band ?CTRL_BITS =/= 0 ->
-    whpan(0.0,0.5);
-generic_event(#mousebutton{button=5,mod=Mod,state=?SDL_RELEASED}, _Camera, _Redraw)
-  when Mod band ?CTRL_BITS =/= 0 ->
-    whpan(0.0,-0.5);
-
-generic_event(#mousebutton{button=4,mod=Mod,state=?SDL_RELEASED}, _Camera, _Redraw)
-  when Mod band ?ALT_BITS =/= 0 ->
-  	zoom_step_alt(-1);
-generic_event(#mousebutton{button=4,state=?SDL_RELEASED}, _Camera, _Redraw) ->
-    zoom_step(-1);
-generic_event(#mousebutton{button=5,mod=Mod,state=?SDL_RELEASED}, _Camera, _Redraw)
-  when Mod band ?ALT_BITS =/= 0 ->
-  	zoom_step_alt(1);
-generic_event(#mousebutton{button=5,state=?SDL_RELEASED}, _Camera, _Redraw) ->
-    zoom_step(1);
+generic_event(#mousewheel{dir=ver, wheel=N, mod=Mod}, Camera, Redraw) ->
+    %% ?dbg("Wheel: ~w  shift: ~p   alt ~p   ctrl: ~p~n",
+    %%      [N,Mod band ?SHIFT_BITS =/= 0,Mod band ?ALT_BITS =/= 0, Mod band ?CTRL_BITS =/= 0]),
+    if Mod band ?SHIFT_BITS =/= 0 andalso Mod band ?ALT_BITS =/= 0 ->
+            whrotate(N/2.0,0.0);
+       Mod band ?SHIFT_BITS =/= 0 ->
+            whrotate(0.0,N/2.0);
+       Mod band ?CTRL_BITS =/= 0 andalso Mod band ?ALT_BITS =/= 0 ->
+            whpan(N/20.0,0.0);
+       Mod band ?CTRL_BITS =/= 0 ->
+            whpan(0.0,N/20.0);
+       Mod band ?ALT_BITS =/= 0 ->
+            zoom_step_alt(-N);
+       is_record(Camera, st), Redraw =:= none ->
+            %% Matching 'none' stops zoom aim from being activated during a drag sequence.
+            %% Zoom aim warps the mouse to the screen's centre, and this can cause a crash
+            %% in since drag events also depend on cursor position.
+            aim_zoom(-N, Camera);
+       true ->
+            zoom_step(-N)
+    end;
+generic_event(#mousewheel{dir=hor, wheel=N, mod=Mod}, _Camera, _Redraw) ->
+    %% ?dbg("Wheel: ~w  shift: ~p   alt ~p   ctrl: ~p~n",
+    %%      [N,Mod band ?SHIFT_BITS =/= 0,Mod band ?ALT_BITS =/= 0, Mod band ?CTRL_BITS =/= 0]),
+    if
+        Mod band ?CTRL_BITS =/= 0 andalso Mod band ?ALT_BITS =/= 0 ->
+            whpan(0.0,N/20.0);
+        Mod band ?CTRL_BITS =/= 0 ->
+            whpan(N/20.0,0.0);
+        Mod band ?ALT_BITS =/= 0 ->
+            whrotate(0.0,N/2.0);
+        true ->  %% rotate X wether SHIFT is pressed or not
+            do_whrotate(N/2.0,0.0)
+    end;
+generic_event(grab_lost, Camera, _Redraw) ->
+    stop_camera(Camera);
 generic_event(_, _, _) -> keep.
 
+aim_zoom(Dir, St0) ->
+    case wings_pref:get_value(highlight_zoom_aim) of
+      true ->
+        #view{origin=OriginB}=Before = wings_view:current(),
+        {{_,Cmd},_} = wings:highlight_aim_setup(St0),
+        wings_view:command(Cmd,St0),
+        #view{origin=OriginA} = wings_view:current(),
+        O = e3d_vec:zero(),
+        if OriginA =:= O, Cmd =:= aim ->
+              wings_view:set_current(Before),
+              zoom_step(Dir);
+            OriginA =:= OriginB ->
+              zoom_step(Dir);
+            true ->
+              Client = wings_wm:this(),
+              {X0,Y0} = wings_wm:win_size(Client),
+              {X,Y} = {X0 div 2, Y0 div 2},
+              wings_io:warp(X,Y),
+              zoom_step(Dir)
+        end;
+      false ->
+        zoom_step(Dir)
+    end.
+
 rotate(Dx, Dy) ->
     Speed = wings_pref:get_value(cam_rotation_speed,25)/25,
     case allow_rotation() of
@@ -756,51 +779,50 @@ whrotate(Dx, Dy) ->
       true ->
         case wings_pref:get_value(wheel_adds, true) of
           false -> keep;
-          true ->
-             wings_wm:dirty(),
-             View0= wings_view:current(),
-             #view{azimuth=Az0,elevation=El0} = View0,
-             S = 2 * wings_pref:get_value(wh_rot_spd),
-             Az = Az0 + Dx*S,
-             El = El0 + Dy*S,
-             View = View0#view{azimuth=Az,elevation=El,along_axis=none},
-             wings_view:set_current(View),
-             keep
+          true -> do_whrotate(Dx,Dy)
         end
     end.
 
+do_whrotate(Dx,Dy) ->
+    case allow_rotation() of
+	false ->
+            keep;
+        true ->
+            wings_wm:dirty(),
+            View0= wings_view:current(),
+            #view{azimuth=Az0,elevation=El0} = View0,
+            S = 2 * wings_pref:get_value(wh_rot_spd),
+            Az = Az0 + Dx*S,
+            El = El0 + Dy*S,
+            View = View0#view{azimuth=Az,elevation=El,along_axis=none},
+            wings_view:set_current(View),
+            keep
+    end.
+
 zoom_step_alt(Dir) ->
     case wings_pref:get_value(wheel_zooms, true) of
 	false -> keep;
 	true ->
-	    wings_wm:dirty(),
-	    #view{distance=Dist} = View = wings_view:current(),
-	    ZoomPercent0 = wings_pref:get_value(wheel_zoom_factor_alt, ?ZOOM_FACTOR_ALT)/100,
-	    ZoomPercent = case wings_pref:get_value(inverted_wheel_zoom) of
-			      false -> ZoomPercent0;
-			      true -> -ZoomPercent0
-			  end,
-	    Delta = dist_factor(Dist)*Dir*ZoomPercent,
-	    wings_view:set_current(View#view{distance=Dist+Delta}),
-	    keep
+	    ZoomFactor = wings_pref:get_value(wheel_zoom_factor_alt, ?ZOOM_FACTOR_ALT),
+		wheel_zoom(ZoomFactor,Dir)
     end.
 
 zoom_step(Dir) ->
     case wings_pref:get_value(wheel_zooms, true) of
 	false -> keep;
 	true ->
-	    wings_wm:dirty(),
-	    #view{distance=Dist} = View = wings_view:current(),
-	    ZoomPercent0 = wings_pref:get_value(wheel_zoom_factor, ?ZOOM_FACTOR)/100,
-	    ZoomPercent = case wings_pref:get_value(inverted_wheel_zoom) of
-			      false -> ZoomPercent0;
-			      true -> -ZoomPercent0
-			  end,
-	    Delta = dist_factor(Dist)*Dir*ZoomPercent,
-	    wings_view:set_current(View#view{distance=Dist+Delta}),
-	    keep
+	    ZoomFactor = wings_pref:get_value(wheel_zoom_factor, ?ZOOM_FACTOR),
+		wheel_zoom(ZoomFactor,Dir)
     end.
 
+wheel_zoom(Factor, Dir) ->
+    wings_wm:dirty(),
+    #view{distance=Dist} = View = wings_view:current(),
+    ZoomPercent = Factor/100,
+    Delta = dist_factor(Dist)*Dir*ZoomPercent,
+    wings_view:set_current(View#view{distance=Dist+Delta}),
+    keep.
+
 zoom(Delta0) ->
     #view{distance=Dist} = View = wings_view:current(),
     Delta = dist_factor(Dist)*Delta0/80,
@@ -815,22 +837,12 @@ pan(Dx0, Dy0) ->
     PanY = PanY0 - Dy,
     wings_view:set_current(View#view{pan_x=PanX,pan_y=PanY}).
 
-arrow_key_pan(Key) when Key=:=?SDLK_LEFT; Key=:=?SDLK_RIGHT; Key=:=?SDLK_UP; Key=:=?SDLK_DOWN ->
-    case wings_pref:get_value(pan_using_arrow_keys) of
-	  true -> arrow_key_pan_1(Key);
-	  _other -> next
-	end;
+arrow_key_pan(?SDLK_LEFT) -> arrow_key_pan(0.05, 0.0);
+arrow_key_pan(?SDLK_RIGHT) -> arrow_key_pan(-0.05, 0.0);
+arrow_key_pan(?SDLK_UP) -> arrow_key_pan(0.0, 0.05);
+arrow_key_pan(?SDLK_DOWN) -> arrow_key_pan(0.0, -0.05);
 arrow_key_pan(_) -> next.
 
-arrow_key_pan_1(?SDLK_LEFT) ->
-    arrow_key_pan(0.25, 0.0);
-arrow_key_pan_1(?SDLK_RIGHT) ->
-    arrow_key_pan(-0.25, 0.0);
-arrow_key_pan_1(?SDLK_UP) ->
-    arrow_key_pan(0.0, 0.25);
-arrow_key_pan_1(?SDLK_DOWN) ->
-    arrow_key_pan(0.0, -0.25).
-
 arrow_key_pan(Dx, Dy) ->
     key_pan(Dx, Dy),
     wings_wm:dirty(),
@@ -838,7 +850,7 @@ arrow_key_pan(Dx, Dy) ->
 
 key_pan(Dx0, Dy0) ->
     #view{pan_x=PanX0,pan_y=PanY0,distance=D} = View = wings_view:current(),
-    S = D/(101-wings_pref:get_value(pan_speed_arrow_keys)),
+    S = D * (wings_pref:get_value(pan_speed_arrow_keys)/100),
     Dx = Dx0*S,
     Dy = Dy0*S,
     PanX = PanX0 + Dx,
@@ -854,7 +866,7 @@ whpan(Dx0, Dy0) ->
     true ->
         wings_wm:dirty(),
         #view{pan_x=PanX0,pan_y=PanY0,distance=D} = View = wings_view:current(),
-        S = D/(101-wings_pref:get_value(wh_pan_spd)),
+        S = D * (wings_pref:get_value(wh_pan_spd)/100),
         Dx = Dx0*S,
         Dy = Dy0*S,
         PanX = PanX0 + Dx,
@@ -864,32 +876,57 @@ whpan(Dx0, Dy0) ->
        end
     end.
 dist_factor(Dist) ->
-    wings_util:max(abs(Dist), 0.2).
+    max(abs(Dist), 0.2).
+
+start_camera(X,Y) ->
+    {0,0,W,H} = wings_wm:viewport(),
+    #camera{x=X,y=Y,ox=X,oy=Y,w=W,h=H}.
 
 stop_camera(#camera{ox=Ox,oy=Oy}) ->
+    wings_wm:release_focus(),
     case wings_io:ungrab(Ox, Oy) of
 	still_grabbed ->
 	    wings_wm:later(view_changed);
 	no_grab ->
-	    wings_wm:release_focus(),
 	    wings_wm:dirty()
     end,
     update_sel(fun show_sel_fun/2),
     pop.
 
-camera_mouse_range(X0, Y0, #camera{x=OX,y=OY, xt=Xt0, yt=Yt0}=Camera) ->
-%%    io:format("Camera Mouse Range ~p ~p~n", [{X0,Y0}, {OX,OY,Xt0,Yt0}]),
-    {X1,Y1} = wings_wm:local2global(X0, Y0),
+camera_mouse_range(X1, Y1, #camera{x=OX,y=OY, xt=Xt0, yt=Yt0}=Camera) ->
+    %% ?dbg("Camera ~w ~w~n", [X1,Y1]),
     XD0 = (X1 - OX),
     YD0 = (Y1 - OY),
     {XD,YD} = wings_pref:lowpass(XD0 + Xt0, YD0 + Yt0),
-
     if
-	XD0 == 0, YD0 == 0 ->
-	    {0.0,0.0,Camera#camera{xt=0,yt=0}};
+	XD0 =:= 0, YD0 =:= 0 ->
+            case wings_pref:get_value(no_warp, false) of
+                false ->
+                    {0.0,0.0,Camera#camera{xt=0,yt=0}};
+                _ ->
+                    #camera{w=W, h=H} = Camera,
+                    Cx = W div 2, Cy = H div 2,
+                    wings_io:warp(Cx, Cy),
+                    {0.0,0.0,Camera#camera{xt=0,yt=0,x=Cx,y=Cy}}
+            end;
 	true ->
-	    wings_io:warp(OX, OY),
-	    {XD/?CAMDIV, YD/?CAMDIV, Camera#camera{xt=XD0, yt=YD0}}
+            case wings_pref:get_value(no_warp, false) of
+                false ->
+                    wings_io:warp(OX, OY),
+                    {XD/?CAMDIV, YD/?CAMDIV, Camera#camera{xt=XD0, yt=YD0}};
+                true ->
+                    %% Warp as few times as possible
+                    #camera{w=W, h=H} = Camera,
+                    case X1 < 10 orelse Y1 < 10 orelse X1 > (W-10) orelse Y1 > (H-10) of
+                        true ->
+                            #camera{w=W, h=H} = Camera,
+                            Cx = W div 2, Cy = H div 2,
+                            wings_io:warp(Cx, Cy),
+                            {XD/?CAMDIV, YD/?CAMDIV, Camera#camera{xt=XD0,yt=YD0,x=Cx,y=Cy}};
+                        false ->
+                            {XD/?CAMDIV, YD/?CAMDIV, Camera#camera{xt=XD0,yt=YD0,x=X1,y=Y1}}
+                    end
+            end
     end.
 
 view_hotkey(Ev, Camera, #state{st=St}) ->
